@@ -21,7 +21,10 @@ use {
             tx_wide_compute_cap, FeatureSet,
         },
         hash::Hash,
-        instruction::{AccountMeta, CompiledInstruction, Instruction, InstructionError},
+        instruction::{
+            AccountMeta, CompiledInstruction, Instruction, InstructionError,
+            TRANSACTION_LEVEL_STACK_HEIGHT,
+        },
         keyed_account::{create_keyed_accounts_unified, keyed_account_at_index, KeyedAccount},
         message::{Message, SanitizedMessage},
         pubkey::Pubkey,
@@ -201,7 +204,7 @@ pub struct InvokeContext<'a> {
     compute_meter: Rc<RefCell<ComputeMeter>>,
     accounts_data_meter: AccountsDataMeter,
     executors: Rc<RefCell<Executors>>,
-    pub instruction_recorder: Option<&'a InstructionRecorder>,
+    pub instruction_trace: Vec<InstructionRecorder>,
     pub feature_set: Arc<FeatureSet>,
     pub timings: ExecuteDetailsTimings,
     pub blockhash: Hash,
@@ -222,7 +225,7 @@ impl<'a> InvokeContext<'a> {
         feature_set: Arc<FeatureSet>,
         blockhash: Hash,
         lamports_per_signature: u64,
-        current_accounts_data_len: u64,
+        initial_accounts_data_len: u64,
     ) -> Self {
         Self {
             invoke_stack: Vec::with_capacity(compute_budget.max_invoke_depth),
@@ -235,9 +238,9 @@ impl<'a> InvokeContext<'a> {
             current_compute_budget: compute_budget,
             compute_budget,
             compute_meter: ComputeMeter::new_ref(compute_budget.max_units),
-            accounts_data_meter: AccountsDataMeter::new(current_accounts_data_len),
+            accounts_data_meter: AccountsDataMeter::new(initial_accounts_data_len),
             executors,
-            instruction_recorder: None,
+            instruction_trace: Vec::new(),
             feature_set,
             timings: ExecuteDetailsTimings::default(),
             blockhash,
@@ -375,8 +378,8 @@ impl<'a> InvokeContext<'a> {
         self.invoke_stack.pop();
     }
 
-    /// Current depth of the invocation stack
-    pub fn invoke_depth(&self) -> usize {
+    /// Current height of the stack
+    pub fn get_stack_height(&self) -> usize {
         self.invoke_stack.len()
     }
 
@@ -439,11 +442,13 @@ impl<'a> InvokeContext<'a> {
                 .checked_add(u128::from(account.lamports()))
                 .ok_or(InstructionError::UnbalancedInstruction)?;
 
+            let pre_data_len = pre_account.data().len() as i64;
+            let post_data_len = account.data().len() as i64;
+            let data_len_delta = post_data_len.saturating_sub(pre_data_len);
             if cap_accounts_data_len {
-                let pre_data_len = pre_account.data().len() as i64;
-                let post_data_len = account.data().len() as i64;
-                let data_len_delta = post_data_len.saturating_sub(pre_data_len);
                 self.accounts_data_meter.consume(data_len_delta)?;
+            } else {
+                self.accounts_data_meter.consume_unchecked(data_len_delta);
             }
 
             Ok(())
@@ -525,11 +530,13 @@ impl<'a> InvokeContext<'a> {
                             pre_account.update(&account);
                         }
 
+                        let pre_data_len = pre_account.data().len() as i64;
+                        let post_data_len = account.data().len() as i64;
+                        let data_len_delta = post_data_len.saturating_sub(pre_data_len);
                         if cap_accounts_data_len {
-                            let pre_data_len = pre_account.data().len() as i64;
-                            let post_data_len = account.data().len() as i64;
-                            let data_len_delta = post_data_len.saturating_sub(pre_data_len);
                             self.accounts_data_meter.consume(data_len_delta)?;
+                        } else {
+                            self.accounts_data_meter.consume_unchecked(data_len_delta);
                         }
 
                         return Ok(());
@@ -566,9 +573,7 @@ impl<'a> InvokeContext<'a> {
             prev_account_sizes.push((account, account_length));
         }
 
-        if let Some(instruction_recorder) = &self.instruction_recorder {
-            instruction_recorder.record_instruction(instruction);
-        }
+        self.record_instruction(self.get_stack_height(), instruction);
 
         let message = SanitizedMessage::Legacy(message);
         self.process_instruction(
@@ -954,6 +959,29 @@ impl<'a> InvokeContext<'a> {
     pub fn get_sysvar_cache(&self) -> &SysvarCache {
         &self.sysvar_cache
     }
+
+    /// Record top-level instruction in the instruction trace
+    pub fn record_top_level_instruction(&mut self, instruction: Instruction) {
+        self.instruction_trace.push(InstructionRecorder::default());
+        self.record_instruction(TRANSACTION_LEVEL_STACK_HEIGHT, instruction);
+    }
+
+    /// Record instruction in the instruction trace
+    pub fn record_instruction(&mut self, stack_height: usize, instruction: Instruction) {
+        if let Some(instruction_recorder) = self.instruction_trace.last() {
+            instruction_recorder.record_instruction(stack_height, instruction)
+        }
+    }
+
+    /// Get the instruction trace
+    pub fn get_instruction_trace(&self) -> &[InstructionRecorder] {
+        &self.instruction_trace
+    }
+
+    /// Get the mutable instruction trace
+    pub fn get_instruction_trace_mut(&mut self) -> &mut Vec<InstructionRecorder> {
+        &mut self.instruction_trace
+    }
 }
 
 pub struct MockInvokeContextPreparation {
@@ -1642,13 +1670,14 @@ mod tests {
         feature_set.deactivate(&requestable_heap_size::id());
         let mut invoke_context = InvokeContext::new_mock(&accounts, &[]);
         invoke_context.feature_set = Arc::new(feature_set);
+        invoke_context.compute_budget = ComputeBudget::new(false);
 
         invoke_context
             .push(&noop_message, &noop_message.instructions()[0], &[0], &[])
             .unwrap();
         assert_eq!(
             *invoke_context.get_compute_budget(),
-            ComputeBudget::default()
+            ComputeBudget::new(false)
         );
         invoke_context.pop();
 
@@ -1658,7 +1687,7 @@ mod tests {
         let expected_compute_budget = ComputeBudget {
             max_units: 500_000,
             heap_size: Some(256_usize.saturating_mul(1024)),
-            ..ComputeBudget::default()
+            ..ComputeBudget::new(false)
         };
         assert_eq!(
             *invoke_context.get_compute_budget(),
@@ -1671,7 +1700,7 @@ mod tests {
             .unwrap();
         assert_eq!(
             *invoke_context.get_compute_budget(),
-            ComputeBudget::default()
+            ComputeBudget::new(false)
         );
         invoke_context.pop();
     }
@@ -1794,7 +1823,7 @@ mod tests {
         let mut invoke_context = InvokeContext::new_mock(&accounts, &builtin_programs);
         invoke_context
             .accounts_data_meter
-            .set_current(user_account_data_len as u64);
+            .set_initial(user_account_data_len as u64);
         invoke_context
             .accounts_data_meter
             .set_maximum(user_account_data_len as u64 * 3);
